uniform vec4 rotCol0;
uniform vec4 rotCol1;
uniform vec4 rotCol2;
uniform vec4 rotCol3;
uniform float zPosition;
uniform vec3 lightPos;

varying vec4 pos;
varying float shadowValue;

bool isInsideCloud(vec3 P);
const float stepsize = 0.05;
const float shadowstep = 0.015;

void main(void) {
	// The position given by gl_Vertex is local to each plane, which means that
	// the z-component is always zero. Because of this, the program is set up
	// to pass the correct z-value for each plane. Here the full 3D position
	// is built, to be used in the fragment shader.
	pos = vec4(gl_Vertex.xy, zPosition, 1.0);
	
	// The actual position of the point (gl_Position) is not altered because
	// I want the planes to be parallel to the view plane at all times. The
	// variable pos is rotated according to the mouse movements of the user
	// though, and this position is used by the fragment shader. In this way,
	// the user can use the mouse to rotate the actual volume while the
	// transparent planes stay in place.
	pos = mat4(rotCol0, rotCol1, rotCol2, rotCol3) * pos;
	
	// Calculate self-shadowing. First a check is performed to see if the point
	// lies within the cloud (as defined by the implicit primitives). If not,
	// the shadowValue is set to 0.0 (the point will be fully transparent
	// anyway). Otherwise, a ray from the point to the lightsource is examined,
	// and the portion of it that is within the cloud volume is calculated.
	// The shadow value lies between 0.0 and 1.0, with 1.0 meaning "completely
	// dark".
	shadowValue = 0.0;
	if (isInsideCloud(pos.xyz)) {
		vec3 dir = lightPos - pos.xyz;
		int steps = int(floor(length(dir)/stepsize));
		dir = normalize(dir);
		for (int i = 1; i <= steps; i++) {
			if (isInsideCloud(pos.xyz + dir*float(i)*stepsize)) {
				shadowValue += shadowstep;
			}
		}
		clamp(shadowValue, 0.0, 1.0);
	}
	
	gl_Position = ftransform();
}

/*
 * bool isInsideCloud(vec3 P)
 *
 * Returns true if the supplied point lies within the implicit primitives 
 * that define the general shape of the cloud and false otherwise.
 */
bool isInsideCloud(vec3 P) {
	// SCENE 1. The code below results in a regular looking cute cloud made out 
	// of ten implicit spheres.
	float sphereValue0 = P.x*P.x + P.y*P.y + P.z*P.z;
	float sphereValue1 = (P.x-0.5)*(P.x-0.5) + P.y*P.y + P.z*P.z;
	float sphereValue2 = (P.x+0.5)*(P.x+0.5) + (P.y+0.2)*(P.y+0.2) + P.z*P.z;
	float sphereValue3 = (P.x-0.3)*(P.x-0.3) + (P.y-0.5)*(P.y-0.5) + P.z*P.z;
	float sphereValue4 = (P.x-0.3)*(P.x-0.3) + (P.y+0.1)*(P.y+0.1) + (P.z-0.5)*(P.z-0.5);
	float sphereValue5 = (P.x+0.3)*(P.x+0.3) + (P.y-0.2)*(P.y-0.2) + P.z*P.z;
	float sphereValue6 = (P.x+0.8)*(P.x+0.8) + (P.y+0.3)*(P.y+0.3) + P.z*P.z;
	float sphereValue7 = (P.x-0.8)*(P.x-0.8) + (P.y+0.2)*(P.y+0.2) + P.z*P.z;
	float sphereValue8 = (P.x-0.3)*(P.x-0.3) + P.y*P.y + (P.z+0.5)*(P.z+0.5);
	float sphereValue9 = (P.x-0.1)*(P.x-0.1) + P.y*P.y + (P.z+0.4)*(P.z+0.4);

	return (sphereValue0 <= 0.4 || 
			sphereValue1 <= 0.5 || 
			sphereValue2 <= 0.3 ||
			sphereValue3 <= 0.3 ||
			sphereValue4 <= 0.3 ||
			sphereValue5 <= 0.25 ||
			sphereValue6 <= 0.2 ||
			sphereValue7 <= 0.2 ||
			sphereValue8 <= 0.3 ||
			sphereValue9 <= 0.2);
	
	// SCENE 2. The code below results in a flat cloud layer in the bottom of 
	// the volume and a small spherical cloud that casts a shadow.
	/*
	float floorValue = length(-1.0-P.y);
	float sphereValue = (P.x+0.3)*(P.x+0.3) + (P.y-0.3)*(P.y-0.3) + (P.z+0.3)*(P.z+0.3);
	return (floorValue <= 0.3 || sphereValue <= 0.2);
	*/
}
